/*
 * sgfstrip: strip away some ocurrences of some properties
 * This works for arbitrary games written in SGF format:
 * the fields are not interpreted.
 *
 * Call: sgfstrip [options] tag1 tag2 .. < input > output
 *
 * Property ids tag1, tag2, ... and empty nodes are removed.
 * Leading and trailing whitespace for all property values is removed.
 *
 * By default, the tags are removed everywhere.
 * -pe: preserve empty nodes
 * -pw: preserve whitespace
 * -h, -m, -t: don't do the action in head / middle / tail
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include "readsgf.h"
#include "writesgf.h"
#include "errexit.h"

FILE *outf;

#define MAXM	1000
char *(propids[MAXM]);
int propidct;

static int should_strip(char *s) {
	int i;

	for (i=0; i<propidct; i++)
		if (!strcmp(s, propids[i]))
			return 1;
	return 0;
}

int optpw, optpe, opth, optm, optt;

static int is_whitespace(char c) {
	return (c == ' ' || c == '\t' || c == '\r' || c == '\n');
}

static void remove_whitespace(char **s) {
	char *t = *s;

	while (is_whitespace(*t))
		t++;
	*s = t;
	while (*t)
		t++;
	while (t > *s && is_whitespace(t[-1]))
		t--;
	*t = 0;
}

int eof;
int peekc;

int gtlevel;
int nodect;
int invariation;

static void
strip_propvalues(struct propvalue *p, int action) {
	while (p) {
		if (action && !optpw)
			remove_whitespace(&(p->val));
		p = p->next;
	}
}

static int
is_final_node(struct node *n) {
	return !invariation && (n->next == 0);
}

static void
strip_property_sequence(struct property **q, int action) {
	struct property *p;

	while (*q) {
		p = *q;

//		int single = (p->val->next == NULL);
//		int empty = (p->val->val[0] == 0);

		if (action && should_strip(p->id)) {
			*q = p->next;
		} else {
			strip_propvalues(p->val, action);
			q = &(p->next);
		}
	}
}

static void
strip_nodesequence(struct node **nn) {
	struct node *n;
	struct property **q;
	int action;

	while (*nn) {
		n = *nn;
		action = ((nodect == 0) ? !opth :
			  (is_final_node(n) ? !optt : !optm));
		q = &(n->p);
		if (*q)
			strip_property_sequence(q, action);
		if (action && !optpe && n->p == NULL) {
			*nn = n->next;
		} else {
			nn = &(n->next);
			nodect++;
		}
	}
}

/* forward declaration */
static void strip_gametree_sequence(struct gametree *g);

static void
strip_gametree(struct gametree *g) {
	int newgame;

	gtlevel++;
	newgame = (gtlevel == 1);
	if (newgame)
		nodect = invariation = 0;
	strip_nodesequence(&(g->nodesequence));
	strip_gametree_sequence(g->firstchild);
	gtlevel--;
	invariation = gtlevel;
}

static void
strip_gametree_sequence(struct gametree *g) {
	while (g) {
		strip_gametree(g);
		g = g->nextsibling;
	}
}

static int
file_exists(char *fn) {
	struct stat sb;

	return (stat(fn, &sb) == 0);
}

static int is_upper_case(char c) {
	return c >= 'A' && c <= 'Z';
}

static int
is_all_caps(char *s) {
	while (*s)
		if (!is_upper_case(*s++))
			return 0;
	return 1;
}

static void
usage() {
	fprintf(stderr, "Usage: "
		"sgfstrip [-h] [-m] [-t] [-pw] [-pe] tags "
		"< input > output\n");
	exit(0);
}

int
main(int argc, char **argv){
	int i;
	char *infile = NULL;
	struct gametree *g;

	progname = "sgfstrip";
	optpw = optpe = 0;

	for (i=1; i<argc; i++) {
		if (!strcmp(argv[i], "-?") || !strcmp(argv[i], "--help"))
			usage();
		if (!strcmp(argv[i], "-h")) {
			opth = 1;
			continue;
		}
		if (!strcmp(argv[i], "-m")) {
			optm = 1;
			continue;
		}
		if (!strcmp(argv[i], "-t")) {
			optt = 1;
			continue;
		}
		if (!strcmp(argv[i], "-pe")) {
			optpe = 1;
			continue;
		}
		if (!strcmp(argv[i], "-pw")) {
			optpw = 1;
			continue;
		}
		if (is_all_caps(argv[i])) {
			if (propidct == 1000)
				errexit("too many args");
			propids[propidct++] = argv[i];
			continue;
		}
		if (argv[i][0] == '-')
			errexit("unrecognized option %s", argv[i]);

		/* maybe input file? */
		if (!infile && file_exists(argv[i])) {
			infile = argv[i];
			continue;
		}

		errexit("unrecognized parameter %s - not all caps", argv[i]);
	}

	outf = stdout;

	if (infile) {
		FILE *f = freopen(infile, "r", stdin);
		if (!f)
			errexit("cannot open %s", infile);
	}

	eof = 0;
	readsgf(NULL, &g);
	gtlevel = 0;
	strip_gametree_sequence(g);
	writesgf(g,stdout);

	return 0;
}
